%% Step 0: Initialize matrices (one time calculation)
function [Q, Qinv, Z, init_vars] = Initialize_TIE_Solver(use_gpu, ...
    use_single, check_answer, plot_data, save_data, rows, columns, ...
    reg_factor1, reg_factor2, Neumann)

% Step 0a: define T matrix
n = min(rows, columns);
r = 0*(1:n); r(1) = -2; r(2) = 1; T = toeplitz(r);
Z = zeros(n);
if use_single
    T = single(T);
    Z = single(Z);
end
if Neumann
    T(1,1) = -1; T(n,n) = -1;
end

% Step 0b: eigenvalue decomposition of T (eigenvectors: Q, eigenvalues: S)
[Q,S] = eig(T);

% Step 0c: T is Hermitian so Q is unitary: Q inverse = Q conj transpose
Qinv = Q'; 

% Step 0d: compute elements of Z (denominator of Eq. 7 with regularization)
% We will pre-calculate Z and use it to find K according to: K = H ./ Z
for i = 1:n
    for j = 1:n
        Z(i,j) = 1/(S(i,i)+S(j,j)) * ...
            ((S(i,i)+S(j,j))^2/((S(i,i)+S(j,j))^2+reg_factor1)); %
    end
end

if use_gpu
    gd = gpuDevice();
    Q = gpuArray(Q);
    Qinv = gpuArray(Qinv);
    Z = gpuArray(Z);
end

% Step 0e: compute the crop window for future images
if rows<columns
    crop_cols = 1+(columns-rows)/2:(columns+rows)/2;
    crop_rows = 1:rows;
else
    crop_rows = 1+(rows-columns)/2:(rows+columns)/2;
    crop_cols = 1:columns;
end

% Step 0f: store intermediate variables for reuse
init_vars.use_gpu = use_gpu;
init_vars.use_single = use_single;
init_vars.crop_cols = crop_cols;
init_vars.crop_rows = crop_rows;
init_vars.reg_factor1 = reg_factor1;
init_vars.reg_factor2 = reg_factor2;
init_vars.check_answer = check_answer;
init_vars.plot_data = plot_data;
init_vars.save_data = save_data;
end